/*!
** @file
** @brief   The customization of the standard Windows controls.
** @remarks Application-specific features are added to the some controls.
** @author  UETA, Shin-ichi <ueta@usdesign.jp>
** @license The MIT License
** @verbatim

Copyright (C) 2014 Ueta System Design Studio

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

@endverbatim
*/
#pragma once

/*!
** @brief   The display monitor list based on the Combo Box control.
** @remarks -
*/
class CMonitorList : public CWindowImpl<CMonitorList, CComboBox>
{
public:

	/*!
	** @brief   Initializes a new CMonitorList object.
	** @remarks -
	*/
	CMonitorList()
	{
		;
	}

	/*!
	** @brief   Subclasses a Combo Box control and attaches it to the object.
	** @param   [in] hWnd a handle to the Combo Box control to be subclassed.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL SubclassWindow(HWND hWnd)
	{
		if (!CWindowImpl<CMonitorList, CComboBox>::SubclassWindow(hWnd))
			return FALSE;

		;

		return TRUE;
	}

	/*!
	** @brief   Appends an item to the Combo Box control.
	** @param   [in] hMonitor a handle to the display monitor.
	** @param   [in] lpszDeviceString the description of the display monitor.
	** @returns The index of the item if successful, or CB_ERR otherwise.
	** @remarks -
	*/
	int AddItem(HMONITOR hMonitor, LPCTSTR lpszDeviceString)
	{
		CString strItem;
		ATLASSERT(lpszDeviceString);
		strItem.Format(TEXT("%d. %s"), GetCount() + 1, lpszDeviceString);
		int iItem = AddString(strItem);

		if (iItem != CB_ERR)
		{
			ATLASSERT(LPARAM(hMonitor));
			ATLVERIFY(SetItemData(iItem, LPARAM(hMonitor)) != CB_ERR);
		}
		return iItem;
	}

	/*!
	** @brief   Searches for an item in the Combo Box control with the specified handle.
	** @param   [in] hMonitor a handle to the display monitor.
	** @returns The index of the item if successful, or CB_ERR otherwise.
	** @remarks -
	*/
	int FindItem(HMONITOR hMonitor) const
	{
		int nItems = GetCount();

		for (int iItem = 0; iItem < nItems; ++iItem)
		{
			if (GetItem(iItem) == hMonitor)
			{
				return iItem;
			}
		}
		return CB_ERR;
	}

	/*!
	** @brief   Retrieves the handle associated with the specified item in the Combo Box control.
	** @param   [in] iItem the index of the item.
	** @returns The handle if successful, or NULL otherwise.
	** @remarks -
	*/
	HMONITOR GetItem(int iItem) const
	{
		LPARAM lParam = CComboBox::GetItemData(iItem);

		if (lParam != CB_ERR)
		{
			HMONITOR hMonitor = HMONITOR(lParam);
			ATLASSERT(hMonitor);
			return hMonitor;
		}
		return NULL;
	}

	/*!
	** @brief   Selects the item in the Combo Box control that has the specified handle.
	** @param   [in] hMonitor a handle to the display monitor.
	** @returns The index of the currently selected item if successful, or CB_ERR otherwise.
	** @remarks -
	*/
	int SetCurSel(HMONITOR hMonitor)
	{
		int iItem = FindItem(hMonitor);

		return CComboBox::SetCurSel(iItem);
	}

	using CComboBox::SetCurSel;

	DECLARE_EMPTY_MSG_MAP()
};

/*!
** @brief   The top-level window list based on the List-View control.
** @remarks -
*/
class CWindowList : public CSortListViewCtrl
{
public:

	/*!
	** @brief   Initializes a new CWindowList object.
	** @remarks -
	*/
	CWindowList()
	{
		;
	}

	/*!
	** @brief   Subclasses a List-View control and attaches it to the object.
	** @param   [in] hWnd a handle to the List-View control to be subclassed.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL SubclassWindow(HWND hWnd)
	{
		if (!CSortListViewCtrl::SubclassWindow(hWnd))
			return FALSE;

		SetExtendedListViewStyle(LVS_EX_FULLROWSELECT);

		CString strHeaders[5];

		for (int iCol = 0; iCol < _countof(strHeaders); ++iCol)
		{
			ATLVERIFY(strHeaders[iCol].LoadString(IDS_HEADER + iCol));
		}
		ATLVERIFY(0 == InsertColumn(0, strHeaders[0], LVCFMT_LEFT | LVCFMT_IMAGE));

		ATLVERIFY(1 == InsertColumn(1, strHeaders[1], LVCFMT_RIGHT));
		ATLVERIFY(2 == InsertColumn(2, strHeaders[2], LVCFMT_RIGHT));
		ATLVERIFY(3 == InsertColumn(3, strHeaders[3], LVCFMT_RIGHT));
		ATLVERIFY(4 == InsertColumn(4, strHeaders[4], LVCFMT_RIGHT));

		SetColumnSortType(0, LVCOLSORT_TEXT);

		SetColumnSortType(1, LVCOLSORT_LONG);
		SetColumnSortType(2, LVCOLSORT_LONG);
		SetColumnSortType(3, LVCOLSORT_LONG);
		SetColumnSortType(4, LVCOLSORT_LONG);

		SetSortColumn(0);

		CRect rcClient;
		ATLVERIFY(GetClientRect(rcClient));
		CRect rcCell(0, 0, 4, 8); // dialog box units
		ATLVERIFY(::MapDialogRect(GetParent(), rcCell));
		int cxVScroll = ::GetSystemMetrics(SM_CXVSCROLL);

		ATLVERIFY(SetColumnWidth(0, rcClient.Width() - rcCell.Width() * 32 - cxVScroll));

		ATLVERIFY(SetColumnWidth(1, rcCell.Width() * 8));
		ATLVERIFY(SetColumnWidth(2, rcCell.Width() * 8));
		ATLVERIFY(SetColumnWidth(3, rcCell.Width() * 8));
		ATLVERIFY(SetColumnWidth(4, rcCell.Width() * 8));

		int cxIcon = ::GetSystemMetrics(SM_CYSMICON);
		int cyIcon = ::GetSystemMetrics(SM_CYSMICON);

		CImageList hImageList;
		ATLVERIFY(hImageList.Create(cxIcon, cyIcon, ILC_COLOR32 | ILC_MASK, 0, 10));
		ATLASSERT(hImageList);
		SetImageList(hImageList, LVSIL_SMALL);

		return TRUE;
	}

	/*!
	** @brief   Deletes all items from the List-View control.
	** @returns No return value.
	** @remarks -
	*/
	void DeleteAllItems()
	{
		CSortListViewCtrl::DeleteAllItems();

		CImageList hImageList = GetImageList(LVSIL_SMALL);
		ATLASSERT(hImageList);
		ATLVERIFY(hImageList.RemoveAll());
	}

	/*!
	** @brief   Appends an item to the List-View control.
	** @param   [in] hWnd a handle to the window.
	** @param   [in] lpRect the bounding rectangle of the window, in screen coordinates.
	** @returns The index of the item if successful, or -1 otherwise.
	** @remarks -
	*/
	int AddItem(HWND hWnd, LPCRECT lpRect)
	{
		CString strWindowText;
		CWindow(hWnd).GetWindowText(strWindowText);
		HICON hIconSm = CWindow(hWnd).GetIcon(FALSE);
		if (!hIconSm)
			hIconSm = HICON(::GetClassLongPtr(hWnd, GCLP_HICONSM));

		CImageList hImageList = GetImageList(LVSIL_SMALL);
		ATLASSERT(hImageList);

		int iImage = I_IMAGENONE;

		if (hIconSm)
		{
			iImage = hImageList.AddIcon(hIconSm);

			if (iImage < 0)
				iImage = I_IMAGENONE;
		}
		int iItem = GetItemCount();

		iItem = CSortListViewCtrl::AddItem(iItem, 0, strWindowText, iImage);

		if (iItem != -1)
		{
			ATLASSERT(LPARAM(hWnd));
			ATLVERIFY(SetItemData(iItem, LPARAM(hWnd)));
			ATLVERIFY(SetSubItems(iItem, lpRect));
		}
		return iItem;
	}

	/*!
	** @brief   Sets the values of the subitems associated with the specified item.
	** @param   [in] iItem the index of the item.
	** @param   [in] lpRect the bounding rectangle of the window, in screen coordinates.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL SetSubItems(int iItem, LPCRECT lpRect)
	{
		CRect rect = lpRect;

		CString strSubItems[5];

		strSubItems[1].Format(TEXT("%d"), rect.Width());
		strSubItems[2].Format(TEXT("%d"), rect.Height());
		strSubItems[3].Format(TEXT("%d"), rect.left);
		strSubItems[4].Format(TEXT("%d"), rect.top);

		BOOL bSuccess = TRUE;

		for (int iSubItem = 1; iSubItem < _countof(strSubItems); ++iSubItem)
		{
			LVITEM item = {LVIF_TEXT};

			item.iItem    = iItem;
			item.iSubItem = iSubItem;
			item.pszText  = strSubItems[iSubItem].GetBuffer();

			if (!CSortListViewCtrl::SetItem(&item))
			{
				bSuccess = FALSE;
			}
		}
		return bSuccess;
	}

	/*!
	** @brief   Searches for an item in the List-View control with the specified handle.
	** @param   [in] hWnd a handle to the window.
	** @returns The index of the item if successful, or -1 otherwise.
	** @remarks -
	*/
	int FindItem(HWND hWnd) const
	{
		LVFINDINFO lvfi = {LVFI_PARAM};

		lvfi.lParam = LPARAM(hWnd);

		int iItem = CSortListViewCtrl::FindItem(&lvfi);

		return iItem;
	}

	/*!
	** @brief   Retrieves the handle associated with the specified item in the List-View control.
	** @param   [in] iItem the index of the item.
	** @returns The handle if successful, or NULL otherwise.
	** @remarks -
	*/
	HWND GetItem(int iItem) const
	{
		LVITEM item = {LVIF_PARAM};

		item.iItem = iItem;

		if (CSortListViewCtrl::GetItem(&item))
		{
			HWND hWnd = HWND(item.lParam);
			ATLASSERT(hWnd);
			return hWnd;
		}
		return NULL;
	}

	/*!
	** @brief   Selects the specified item in the List-View control.
	** @param   [in] iItem the index of the item, or -1 to have no selection.
	** @returns The index of the currently selected item if successful, or -1 otherwise.
	** @remarks -
	*/
	int SelectItem(int iItem)
	{
		if (iItem != -1)
		{
			if (CSortListViewCtrl::SelectItem(iItem))
			{
				return iItem;
			}
		}
		else // deselect
		{
			ATLVERIFY(SetItemState(-1, 0, LVIS_SELECTED));
		}
		return -1;
	}

	/*!
	** @brief   Selects the item in the List-View control that has the specified handle.
	** @param   [in] hWnd a handle to the window.
	** @returns The index of the currently selected item if successful, or -1 otherwise.
	** @remarks -
	*/
	int SelectItem(HWND hWnd)
	{
		int iItem = FindItem(hWnd);

		return SelectItem(iItem);
	}

	/*!
	** @brief   Sorts the items in the List-View control.
	** @param   [in] iCol the index of the column on which the items are sorted.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL SortItems(int iCol = -1)
	{
		if (iCol == -1)
		{
			iCol = GetSortColumn();
		}
		bool bDescending = (iCol == GetSortColumn()) ? IsSortDescending() : false;

		return DoSortItems(iCol, bDescending);
	}
};

/*!
** @brief   The coordinate value editor based on the Combo Box control.
** @remarks -
*/
class CCoordValue : public CWindowImpl<CCoordValue, CComboBox>
{
public:

	static const int MIN = SHRT_MIN; ///< The minimum value for the coordinate value.
	static const int MAX = SHRT_MAX; ///< The maximum value for the coordinate value.

private:

	CContainedWindowT<CEdit> m_wndItem;

	CUpDownCtrl m_wndUpDownCtrl;

	CString m_strItem;

public:

	/*!
	** @brief   Initializes a new CCoordValue object.
	** @remarks -
	*/
	CCoordValue() : m_wndItem(this, 1)
	{
		;
	}

	/*!
	** @brief   Subclasses a Combo Box control and attaches it to the object.
	** @param   [in] hWnd a handle to the Combo Box control to be subclassed.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL SubclassWindow(HWND hWnd)
	{
		if (!CWindowImpl<CCoordValue, CComboBox>::SubclassWindow(hWnd))
			return FALSE;

		COMBOBOXINFO cbi = {sizeof(COMBOBOXINFO)};
		ATLVERIFY(GetComboBoxInfo(&cbi));
		ATLASSERT(::IsWindow(cbi.hwndItem));
		m_wndItem.SubclassWindow(cbi.hwndItem);

		m_wndUpDownCtrl = GetWindow(GW_HWNDNEXT);
		ATLASSERT(::IsWindow(m_wndUpDownCtrl));
		m_wndUpDownCtrl.SetRange(MIN, MAX);

		CRect rcComboBox;
		ATLVERIFY(GetWindowRect(rcComboBox));
		CRect rcUpDownCtrl;
		ATLVERIFY(m_wndUpDownCtrl.GetWindowRect(rcUpDownCtrl));

		CWindow wndParent = GetParent();

		::MapWindowPoints(HWND_DESKTOP, wndParent, reinterpret_cast<POINT*>(&rcComboBox),   2);
		::MapWindowPoints(HWND_DESKTOP, wndParent, reinterpret_cast<POINT*>(&rcUpDownCtrl), 2);

		rcUpDownCtrl.top    = rcComboBox.top;
		rcUpDownCtrl.bottom = rcComboBox.bottom;

		ATLVERIFY(m_wndUpDownCtrl.MoveWindow(rcUpDownCtrl));

		return TRUE;
	}

	BEGIN_MSG_MAP(CCoordValue)
		MSG_WM_ENABLE(OnEnable)
		REFLECTED_COMMAND_CODE_HANDLER_EX(CBN_DROPDOWN, OnDropDown)
		REFLECTED_COMMAND_CODE_HANDLER_EX(CBN_CLOSEUP,  OnCloseUp)
		REFLECTED_COMMAND_CODE_HANDLER_EX(CBN_SELENDOK, OnSelEndOk)
	ALT_MSG_MAP(1)
		MESSAGE_HANDLER(WM_KEYDOWN, OnKey)
		MESSAGE_HANDLER(WM_KEYUP,   OnKey)
	END_MSG_MAP()

	/*!
	** @brief   Called when the application changes the enabled state of the control.
	** @param   [in] bEnable indicates whether to enable or disable the control.
	** @returns No return value.
	** @remarks -
	*/
	void OnEnable(BOOL bEnable)
	{
		DefWindowProc();

		m_wndUpDownCtrl.EnableWindow(bEnable);
	}

	/*!
	** @brief   Handles the CBN_DROPDOWN notification reflected from the parent window.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the control sending the message.
	** @param   [in] hWndCtrl a handle to the control sending the message.
	** @returns No return value.
	** @remarks -
	*/
	void OnDropDown(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(hWndCtrl);

		m_wndItem.GetWindowText(m_strItem);
	}

	/*!
	** @brief   Handles the CBN_CLOSEUP notification reflected from the parent window.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the control sending the message.
	** @param   [in] hWndCtrl a handle to the control sending the message.
	** @returns No return value.
	** @remarks -
	*/
	void OnCloseUp(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(hWndCtrl);

		m_wndItem.SetWindowText(m_strItem), m_strItem.Empty();

		m_wndItem.SetSelAll();
	}

	/*!
	** @brief   Handles the CBN_SELENDOK notification reflected from the parent window.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the control sending the message.
	** @param   [in] hWndCtrl a handle to the control sending the message.
	** @returns No return value.
	** @remarks -
	*/
	void OnSelEndOk(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(hWndCtrl);

		int iCurSel = GetCurSel();

		::PostMessage(GetParent(), WM_COMMAND, MAKEWPARAM(nID + iCurSel, 0), LPARAM(0));
	}

	/*!
	** @brief   Handles the WM_KEYDOWN/WM_KEYUP message for the edit box.
	** @param   [in] uMsg specifies the message.
	** @param   [in] wParam additional message-specific information.
	** @param   [in] lParam additional message-specific information.
	** @param   [in, out] bHandled set to FALSE if further processing is needed.
	** @returns The result of the message processing.
	** @remarks -
	*/
	LRESULT OnKey(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
	{
		switch (TCHAR(wParam))
		{
		case VK_UP:
		case VK_DOWN:

			if (!GetDroppedState()) // get handled finally by the Up-Down control
			{
				return ::SendMessage(m_hWnd, uMsg, wParam, lParam);
			}
			bHandled = FALSE;
			return LRESULT(0);

		default:

			bHandled = FALSE;
			return LRESULT(0);
		}
	}
};
